# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

__version_info__ = ("0", "1", "1")
__version__ = ".".join(__version_info__)
