// Copyright (c) 2019, Lawrence Livermore National Security, LLC.  
// See top-level LICENSE file for details.

#pragma once

#include "wrapConfigManager.h"

#ifdef __cplusplus
#include "ConfigManager.h"
#include "ChannelController.h"

extern "C" {
#endif

#ifdef __cplusplus
} // extern "C"
#endif
