/*************************************************************************
 * Copyright (c) 2015-2016, NVIDIA CORPORATION. All rights reserved.
 *
 * See LICENSE.txt for license information
 ************************************************************************/

#include "common.h"
#include "cuda_runtime.h"

void
print_header()
{
    PRINT("# %10s  %12s  %6s  %6s            out-of-place                       in-place "
          "         \n",
          "", "", "", "");
    PRINT("# %10s  %12s  %6s  %6s  %7s  %6s  %6s  %5s  %7s  %6s  %6s  %5s\n", "size",
          "count", "type", "root", "time", "algbw", "busbw", "error", "time", "algbw",
          "busbw", "error");
    PRINT("# %10s  %12s  %6s  %6s  %7s  %6s  %6s  %5s  %7s  %6s  %6s  %5s\n", "(B)",
          "(elements)", "", "", "(us)", "(GB/s)", "(GB/s)", "", "(us)", "(GB/s)",
          "(GB/s)", "");
}

void
print_line_header(size_t size, size_t count, const char* typeName, const char* opName,
                  int root)
{
    PRINT("%12li  %12li  %6s  %6i", size, count, typeName, root);
}

void
BroadcastGetCollByteCount(size_t* sendcount, size_t* recvcount, size_t* paramcount,
                          size_t* sendInplaceOffset, size_t* recvInplaceOffset,
                          size_t count, int nranks)
{
    *sendcount         = count;
    *recvcount         = count;
    *sendInplaceOffset = 0;
    *recvInplaceOffset = 0;
    *paramcount        = *sendcount;
}

testResult_t
BroadcastInitData(struct threadArgs* args, ncclDataType_t type, ncclRedOp_t op, int root,
                  int rep, int in_place)
{
    size_t sendcount = args->sendBytes / wordSize(type);
    size_t recvcount = args->expectedBytes / wordSize(type);

    for(int i = 0; i < args->nGpus; i++)
    {
        int gpuid = args->localRank * args->nThreads * args->nGpus +
                    args->thread * args->nGpus + i;
        CUDACHECK(cudaSetDevice(gpuid));
        int rank = ((args->proc * args->nThreads + args->thread) * args->nGpus + i);
        CUDACHECK(cudaMemset(args->recvbuffs[i], 0, args->expectedBytes));
        void* data = in_place ? args->recvbuffs[i] : args->sendbuffs[i];
        if(rank == root)
            TESTCHECK(InitData(data, sendcount, type, rep, rank));
        TESTCHECK(InitData(args->expected[i], recvcount, type, rep, root));
        CUDACHECK(cudaDeviceSynchronize());
    }
    return testSuccess;
}

void
BroadcastGetBw(size_t count, int typesize, double sec, double* algBw, double* busBw,
               int nranks)
{
    double baseBw = (double) (count * typesize) / 1.0E9 / sec;

    *algBw        = baseBw;
    double factor = 1;
    *busBw        = baseBw * factor;
}

testResult_t
BroadcastRunColl(void* sendbuff, void* recvbuff, size_t count, ncclDataType_t type,
                 ncclRedOp_t op, int root, ncclComm_t comm, cudaStream_t stream)
{
    int rank;
    NCCLCHECK(ncclCommUserRank(comm, &rank));
#if NCCL_MAJOR >= 2 && NCCL_MINOR >= 2
    NCCLCHECK(ncclBroadcast(sendbuff, recvbuff, count, type, root, comm, stream));
#else
    if(rank == root)
    {
        NCCLCHECK(ncclBcast(sendbuff, count, type, root, comm, stream));
    }
    else
    {
        NCCLCHECK(ncclBcast(recvbuff, count, type, root, comm, stream));
    }
#endif
    return testSuccess;
}

struct testColl broadcastTest = { "Broadcast", BroadcastGetCollByteCount,
                                  BroadcastInitData, BroadcastGetBw, BroadcastRunColl };

void
BroadcastGetBuffSize(size_t* sendcount, size_t* recvcount, size_t count, int nranks)
{
    size_t paramcount, sendInplaceOffset, recvInplaceOffset;
    BroadcastGetCollByteCount(sendcount, recvcount, &paramcount, &sendInplaceOffset,
                              &recvInplaceOffset, count, nranks);
}

testResult_t
BroadcastRunTest(struct threadArgs* args, int root, ncclDataType_t type,
                 const char* typeName, ncclRedOp_t op, const char* opName)
{
    args->collTest = &broadcastTest;
    ncclDataType_t* run_types;
    const char**    run_typenames;
    int             type_count;
    int             begin_root, end_root;

    if((int) type != -1)
    {
        type_count    = 1;
        run_types     = &type;
        run_typenames = &typeName;
    }
    else
    {
        type_count    = ncclNumTypes;
        run_types     = test_types;
        run_typenames = test_typenames;
    }

    if(root != -1)
    {
        begin_root = end_root = root;
    }
    else
    {
        begin_root = 0;
        end_root   = args->nProcs * args->nThreads * args->nGpus - 1;
    }

    for(int i = 0; i < type_count; i++)
    {
        for(int j = begin_root; j <= end_root; j++)
        {
            TESTCHECK(
                TimeTest(args, run_types[i], run_typenames[i], (ncclRedOp_t) 0, "", j));
        }
    }
    return testSuccess;
}

struct testEngine broadcastEngine = { BroadcastGetBuffSize, BroadcastRunTest };
