// wrapCaliper.h
// This file is generated by Shroud 0.12.2. Do not edit.
/**
 * \file wrapCaliper.h
 * \brief Shroud generated wrapper for cali namespace
 */
// For C users and C++ implementation

#ifndef WRAPCALIPER_H
#define WRAPCALIPER_H

#include "typesCaliper.h"

// splicer begin CXX_declarations
// splicer end CXX_declarations

#ifdef __cplusplus
extern "C" {
#endif

//  cali::cali_attr_properties
enum cali_cali_attr_properties {
    cali_CALI_ATTR_DEFAULT = 0,
    cali_CALI_ATTR_ASVALUE = 1,
    cali_CALI_ATTR_NOMERGE = 2,
    cali_CALI_ATTR_SCOPE_PROCESS = 12,
    cali_CALI_ATTR_SCOPE_THREAD = 20,
    cali_CALI_ATTR_SCOPE_TASK = 24,
    cali_CALI_ATTR_SKIP_EVENTS = 64,
    cali_CALI_ATTR_HIDDEN = 128,
    cali_CALI_ATTR_NESTED = 256,
    cali_CALI_ATTR_GLOBAL = 512,
    cali_CALI_ATTR_UNALIGNED = 1024
};

//  cali::cali_attr_type
enum cali_cali_attr_type {
    cali_CALI_TYPE_INV = 0,
    cali_CALI_TYPE_USR = 1,
    cali_CALI_TYPE_INT = 2,
    cali_CALI_TYPE_UINT = 3,
    cali_CALI_TYPE_STRING = 4,
    cali_CALI_TYPE_ADDR = 5,
    cali_CALI_TYPE_DOUBLE = 6,
    cali_CALI_TYPE_BOOL = 7,
    cali_CALI_TYPE_TYPE = 8,
    cali_CALI_TYPE_PTR = 9
};

//  cali::cali_flush_opts
enum cali_cali_flush_opts {
    cali_CALI_FLUSH_CLEAR_BUFFERS = 1
};

// splicer begin C_declarations
// splicer end C_declarations

#ifdef __cplusplus
}
#endif

#endif  // WRAPCALIPER_H
