/*************************************************************************
 * Copyright (c) 2016-2019, NVIDIA CORPORATION. All rights reserved.
 *
 * See LICENSE.txt for license information
 ************************************************************************/

#include "common.h"
#include "cuda_runtime.h"

void
print_header()
{
    PRINT("# %10s  %12s  %6s  %6s            out-of-place                       in-place "
          "         \n",
          "", "", "", "");
    PRINT("# %10s  %12s  %6s  %6s  %7s  %6s  %6s  %5s  %7s  %6s  %6s  %5s\n", "size",
          "count", "type", "redop", "time", "algbw", "busbw", "error", "time", "algbw",
          "busbw", "error");
    PRINT("# %10s  %12s  %6s  %6s  %7s  %6s  %6s  %5s  %7s  %6s  %6s  %5s\n", "(B)",
          "(elements)", "", "", "(us)", "(GB/s)", "(GB/s)", "", "(us)", "(GB/s)",
          "(GB/s)", "");
}

void
print_line_header(size_t size, size_t count, const char* typeName, const char* opName,
                  int root)
{
    PRINT("%12li  %12li  %6s  %6s", size, count, typeName, opName);
}

void
AllReduceGetCollByteCount(size_t* sendcount, size_t* recvcount, size_t* paramcount,
                          size_t* sendInplaceOffset, size_t* recvInplaceOffset,
                          size_t count, int nranks)
{
    *sendcount         = count;
    *recvcount         = count;
    *sendInplaceOffset = 0;
    *recvInplaceOffset = 0;
    *paramcount        = *sendcount;
}

testResult_t
AllReduceInitData(struct threadArgs* args, ncclDataType_t type, ncclRedOp_t op, int root,
                  int rep, int in_place)
{
    size_t sendcount = args->sendBytes / wordSize(type);
    size_t recvcount = args->expectedBytes / wordSize(type);
    int    nranks    = args->nProcs * args->nThreads * args->nGpus;

    for(int i = 0; i < args->nGpus; i++)
    {
        int gpuid = args->localRank * args->nThreads * args->nGpus +
                    args->thread * args->nGpus + i;
        CUDACHECK(cudaSetDevice(gpuid));
        int rank = ((args->proc * args->nThreads + args->thread) * args->nGpus + i);
        CUDACHECK(cudaMemset(args->recvbuffs[i], 0, args->expectedBytes));
        void* data = in_place ? args->recvbuffs[i] : args->sendbuffs[i];
        TESTCHECK(InitData(data, sendcount, type, rep, rank));
        TESTCHECK(InitDataReduce(args->expected[i], recvcount, 0, type, op, rep, nranks));
        CUDACHECK(cudaDeviceSynchronize());
    }
    return testSuccess;
}

void
AllReduceGetBw(size_t count, int typesize, double sec, double* algBw, double* busBw,
               int nranks)
{
    double baseBw = (double) (count * typesize) / 1.0E9 / sec;

    *algBw        = baseBw;
    double factor = ((double) (2 * (nranks - 1))) / ((double) nranks);
    *busBw        = baseBw * factor;
}

testResult_t
AllReduceRunColl(void* sendbuff, void* recvbuff, size_t count, ncclDataType_t type,
                 ncclRedOp_t op, int root, ncclComm_t comm, cudaStream_t stream)
{
    NCCLCHECK(ncclAllReduce(sendbuff, recvbuff, count, type, op, comm, stream));
    return testSuccess;
}

struct testColl allReduceTest = { "AllReduce", AllReduceGetCollByteCount,
                                  AllReduceInitData, AllReduceGetBw, AllReduceRunColl };

void
AllReduceGetBuffSize(size_t* sendcount, size_t* recvcount, size_t count, int nranks)
{
    size_t paramcount, sendInplaceOffset, recvInplaceOffset;
    AllReduceGetCollByteCount(sendcount, recvcount, &paramcount, &sendInplaceOffset,
                              &recvInplaceOffset, count, nranks);
}

testResult_t
AllReduceRunTest(struct threadArgs* args, int root, ncclDataType_t type,
                 const char* typeName, ncclRedOp_t op, const char* opName)
{
    args->collTest = &allReduceTest;
    ncclDataType_t* run_types;
    ncclRedOp_t*    run_ops;
    const char **   run_typenames, **run_opnames;
    int             type_count, op_count;

    if((int) type != -1)
    {
        type_count    = 1;
        run_types     = &type;
        run_typenames = &typeName;
    }
    else
    {
        type_count    = ncclNumTypes;
        run_types     = test_types;
        run_typenames = test_typenames;
    }

    if((int) op != -1)
    {
        op_count    = 1;
        run_ops     = &op;
        run_opnames = &opName;
    }
    else
    {
        op_count    = ncclNumOps;
        run_ops     = test_ops;
        run_opnames = test_opnames;
    }

    for(int i = 0; i < type_count; i++)
    {
        for(int j = 0; j < op_count; j++)
        {
            TESTCHECK(TimeTest(args, run_types[i], run_typenames[i], run_ops[j],
                               run_opnames[j], -1));
        }
    }
    return testSuccess;
}

struct testEngine allReduceEngine = { AllReduceGetBuffSize, AllReduceRunTest };
