// MIT License
//
// Copyright (c) 2020, The Regents of the University of California,
// through Lawrence Berkeley National Laboratory (subject to receipt of any
// required approvals from the U.S. Dept. of Energy).  All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#pragma once

#include "timemory/components/macros.hpp"
#include "timemory/enum.h"
#include "timemory/macros/os.hpp"
#include "timemory/mpl/types.hpp"

TIMEMORY_DECLARE_COMPONENT(network_stats)
//
TIMEMORY_STATISTICS_TYPE(component::network_stats, TIMEMORY_ESC(std::array<int64_t, 8>))
//
TIMEMORY_SET_COMPONENT_API(component::network_stats, project::timemory,
                           os::supports_linux)
//
TIMEMORY_DEFINE_CONCRETE_TRAIT(sampler, component::network_stats, true_type)
TIMEMORY_DEFINE_CONCRETE_TRAIT(file_sampler, component::network_stats, true_type)
TIMEMORY_DEFINE_CONCRETE_TRAIT(echo_enabled, component::network_stats, false_type)
// TIMEMORY_DEFINE_CONCRETE_TRAIT(permissive_statistics, component::network_stats,
// true_type)
//
#if !defined(_LINUX)
TIMEMORY_DEFINE_CONCRETE_TRAIT(is_available, component::network_stats, false_type)
#endif
//
TIMEMORY_PROPERTY_SPECIALIZATION(network_stats, TIMEMORY_NETWORK_STATS, "network_stats",
                                 "network")
//
// TIMEMORY_METADATA_SPECIALIZATION(network_stats, "network_stats",
//                                 "Reports network bytes, packets, errors, dropped")
